/* *{
    font-size:18px;
} */


/* #TaskWindow{ */
.large{
    font-size:25px;
}

#FuzzyTask{
    font-size:30px;
    padding:12px;

}

#RelevantQuestion{
    font-weight: 100;
    font-size:35px;
    opacity: .3;
}

.bold{
    font-weight: bold;
}

.done{
    text-decoration-line: line-through;
    text-decoration:line-through;
}